import httpx


class ReportGeneratorSyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return {"API-KEY": self.altscore_client.api_key}

    def generate(self, report_request: dict) -> str:
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                "/v1/tools/generate-report",
                headers=self.build_headers(),
                json=report_request,
                timeout=120
            )
            if response.status_code == 200:
                return response.json()["url"]
            else:
                raise Exception(response.text)


class ReportGeneratorAsyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return {"API-KEY": self.altscore_client.api_key}

    async def generate(self, report_request: dict) -> str:
        with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                "/v1/tools/generate-report",
                headers=self.build_headers(),
                json=report_request,
                timeout=120
            )
            if response.status_code == 200:
                return response.json()["url"]
            else:
                raise Exception(response.text)



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class RuleAlert(BaseModel):
    level: int = Field(alias="level")
    message: str = Field(alias="message")


class RuleAPIDTO(BaseModel):
    id: str = Field(alias="id")
    code: str = Field(alias="code")
    label: str = Field(alias="label")
    alerts: List[RuleAlert] = Field(alias="alerts", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateRuleDTO(BaseModel):
    label: str = Field(alias="label")
    code: str = Field(alias="code")
    alerts: List[RuleAlert] = Field(alias="alerts")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class RulesSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "rules", header_builder, renew_token, RuleAPIDTO.parse_obj(data))


class RulesAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "rules", header_builder, renew_token, RuleAPIDTO.parse_obj(data))


class RulesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=RulesSync,
                         retrieve_data_model=RuleAPIDTO,
                         create_data_model=CreateRuleDTO,
                         update_data_model=CreateRuleDTO,
                         resource="rules")


class RulesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=RulesAsync,
                         retrieve_data_model=RuleAPIDTO,
                         create_data_model=CreateRuleDTO,
                         update_data_model=CreateRuleDTO,
                         resource="rules")



from pydantic import BaseModel, Field
from typing import Optional
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401
from altscore.cms.model.credit_account import CreditAccountSync, CreditAccountAsync, CreditAccountAPIDTO
from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule
import datetime as dt


class ClientAPIDTO(BaseModel):
    id: str = Field(alias="clientId")
    partner_id: str = Field(alias="partnerId")
    status: str = Field(alias="status")
    external_id: str = Field(alias="externalId")
    tax_id: str = Field(alias="taxId")
    email_address: str = Field(alias="emailAddress")
    dba: str = Field(alias="dba")
    legal_name: str = Field(alias="legalName")
    address: str = Field(alias="address")
    phone_number: Optional[str] = Field(alias="phoneNumber", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class CreateClientDTO(BaseModel):
    partner_id: str = Field(alias="partnerId")
    external_id: str = Field(alias="externalId")
    legal_name: str = Field(alias="legalName")
    tax_id: str = Field(alias="taxId")
    dba: str = Field(alias="dba")
    address: str = Field(alias="address")
    email_address: str = Field(alias="emailAddress")
    phone_number: Optional[str] = Field(alias="phoneNumber", default="")
    activation_date: Optional[str] = Field(alias="activationDate", default=dt.date.today().strftime("%Y-%m-%d"))

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class ClientBase:

    @staticmethod
    def _credit_accounts(
            client_id: str, product_family: str
    ) -> (str, Optional[dict]):
        return f"/v2/clients/{client_id}/credit-accounts/{product_family}"

    @staticmethod
    def _status(client_id: str):
        return f"/v2/clients/{client_id}/status"


class ClientAsync(ClientBase):
    data: ClientAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: ClientAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401
    async def get_credit_account(self, product_family: str) -> CreditAccountAsync:
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                self._credit_accounts(self.data.id, product_family=product_family),
                headers=self._header_builder()
            )
            return CreditAccountAsync(
                base_url=self.base_url,
                header_builder=self._header_builder,
                renew_token=self.renew_token,
                data=CreditAccountAPIDTO.parse_obj(response.json())
            )

    @retry_on_401
    async def enable(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.put(
                self._status(self.data.id),
                json={
                    "status": "enabled"
                },
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            self.data = ClientAPIDTO.parse_obj(response.json())

    @retry_on_401
    async def disable(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.put(
                self._status(self.data.id),
                json={
                    "status": "disabled"
                },
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            self.data = ClientAPIDTO.parse_obj(response.json())

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class ClientSync(ClientBase):
    data: ClientAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: ClientAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data: ClientAPIDTO = data

    @retry_on_401
    def get_credit_account(self, product_family: str) -> CreditAccountSync:
        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                self._credit_accounts(self.data.id, product_family=product_family),
                headers=self._header_builder()
            )
            return CreditAccountSync(
                base_url=self.base_url,
                header_builder=self._header_builder,
                renew_token=self.renew_token,
                data=CreditAccountAPIDTO.parse_obj(response.json())
            )

    @retry_on_401
    def enable(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.put(
                self._status(self.data.id),
                json={
                    "status": "enabled"
                },
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            self.data = ClientAPIDTO.parse_obj(response.json())

    @retry_on_401
    def disable(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.put(
                self._status(self.data.id),
                json={
                    "status": "disabled"
                },
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            self.data = ClientAPIDTO.parse_obj(response.json())

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class ClientsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=ClientAsync,
            retrieve_data_model=ClientAPIDTO,
            create_data_model=CreateClientDTO,
            update_data_model=None,
            resource="clients",
            resource_version="v2"
        )

    def renew_token(self):
        self.altscore_client.renew_token()

    @retry_on_401
    async def retrieve_by_external_id(self, external_id: str, partner_id: str = None) -> Optional[ClientAsync]:
        headers = self.build_headers()
        if partner_id is not None:
            headers["x-partner-id"] = partner_id

        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            response = await client.get(
                f"/{self.resource_version}/{self.resource}/{external_id}",
                headers=headers,
                timeout=30
            )
            if response.status_code == 200:
                return self.async_resource(
                    base_url=self.altscore_client._cms_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=self.retrieve_data_model.parse_obj(response.json())
                )
            return None

    @retry_on_401
    async def create(self, new_entity_data: dict):
        partner_id = new_entity_data.get("partnerId")
        if partner_id is None:
            partner_id = self.altscore_client.partner_id
            new_entity_data["partnerId"] = partner_id

        headers = self.build_headers()
        # if this is a client creation with "impersonation" of a partner we need to set the x-partner-id header
        headers["x-partner-id"] = partner_id
        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            response = await client.post(
                "/v2/clients",
                headers=headers,
                json=CreateClientDTO.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return response.json()["clientId"]


class ClientsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=ClientSync,
            retrieve_data_model=ClientAPIDTO,
            create_data_model=CreateClientDTO,
            update_data_model=None,
            resource="clients",
            resource_version="v2"
        )

    def renew_token(self):
        self.altscore_client.renew_token()

    @retry_on_401
    def retrieve_by_external_id(self, external_id: str, partner_id: str = None) -> Optional[ClientSync]:

        headers = self.build_headers()
        if partner_id is not None:
            headers["x-partner-id"] = partner_id

        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.get(
                f"/{self.resource_version}/{self.resource}/{external_id}",
                headers=headers,
                timeout=30
            )
            if response.status_code == 200:
                return self.sync_resource(
                    base_url=self.altscore_client._cms_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=self.retrieve_data_model.parse_obj(response.json())
                )
            return None

    @retry_on_401
    def create(self, new_entity_data: dict):
        partner_id = new_entity_data.get("partnerId")
        if partner_id is None:
            partner_id = self.altscore_client.partner_id
            new_entity_data["partnerId"] = partner_id

        headers = self.build_headers()
        # if this is a client creation with "impersonation" of a partner we need to set the x-partner-id header
        headers["x-partner-id"] = partner_id
        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.post(
                "/v2/clients",
                headers=headers,
                json=CreateClientDTO.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return response.json()["clientId"]



from altscore.borrower_central.model.borrower import BorrowersAsyncModule, BorrowersSyncModule
from altscore.borrower_central.model.data_models import DataModelAsyncModule, DataModelSyncModule
from altscore.borrower_central.model.executions import ExecutionAsyncModule, ExecutionSyncModule
from altscore.borrower_central.model.tools import ReportGeneratorAsyncModule, ReportGeneratorSyncModule
from altscore.borrower_central.model.form_templates import FormTemplatesAsyncModule, FormTemplatesSyncModule
from altscore.borrower_central.model.otp_templates import OTPTemplatesAsyncModule, OTPTemplatesSyncModule
from altscore.borrower_central.model.addresses import AddressesAsyncModule, AddressesSyncModule
from altscore.borrower_central.model.authorizations import AuthorizationsAsyncModule, AuthorizationsSyncModule
from altscore.borrower_central.model.borrower_fields import BorrowerFieldsAsyncModule, BorrowerFieldsSyncModule
from altscore.borrower_central.model.documents import DocumentsAsyncModule, DocumentsSyncModule
from altscore.borrower_central.model.identities import IdentitiesAsyncModule, IdentitiesSyncModule
from altscore.borrower_central.model.points_of_contact import PointOfContactAsyncModule, PointOfContactSyncModule
from altscore.borrower_central.model.relationships import RelationshipsAsyncModule, RelationshipsSyncModule
from altscore.borrower_central.model.store_packages import PackagesAsyncModule, PackagesSyncModule
from altscore.borrower_central.model.store_sources import SourcesAsyncModule, SourcesSyncModule
from altscore.borrower_central.model.workflows import WorkflowsAsyncModule, WorkflowsSyncModule
from altscore.borrower_central.model.evaluators import EvaluatorAsyncModule, EvaluatorSyncModule
from altscore.borrower_central.model.forms import FormsAsyncModule, FormsSyncModule
from altscore.borrower_central.model.policy_alerts import AlertsAsyncModule, AlertsSyncModule
from altscore.borrower_central.model.policy_rules import RulesAsyncModule, RulesSyncModule
from altscore.borrower_central.model.integrations.sat import SatIntegrationAsyncModule, SatIntegrationSyncModule


class BorrowerCentralAsync:

    def __init__(self, altscore_client):
        self.addresses = AddressesAsyncModule(altscore_client)
        self.authorizations = AuthorizationsAsyncModule(altscore_client)
        self.borrowers = BorrowersAsyncModule(altscore_client)
        self.borrower_fields = BorrowerFieldsAsyncModule(altscore_client)
        self.data_models = DataModelAsyncModule(altscore_client)
        self.documents = DocumentsAsyncModule(altscore_client)
        self.executions = ExecutionAsyncModule(altscore_client)
        self.identities = IdentitiesAsyncModule(altscore_client)
        self.workflows = WorkflowsAsyncModule(altscore_client)
        self.points_of_contact = PointOfContactAsyncModule(altscore_client)
        self.relationships = RelationshipsAsyncModule(altscore_client)
        self.report_generator = ReportGeneratorAsyncModule(altscore_client)
        self.form_templates = FormTemplatesAsyncModule(altscore_client)
        self.otp_templates = OTPTemplatesAsyncModule(altscore_client)
        self.store_packages = PackagesAsyncModule(altscore_client)
        self.store_sources = SourcesAsyncModule(altscore_client)
        self.evaluators = EvaluatorAsyncModule(altscore_client)
        self.forms = FormsAsyncModule(altscore_client)
        self.alerts = AlertsAsyncModule(altscore_client)
        self.rules = RulesAsyncModule(altscore_client)
        self.sat_integration = SatIntegrationAsyncModule(altscore_client)


class BorrowerCentralSync:

    def __init__(self, altscore_client):
        self.addresses = AddressesSyncModule(altscore_client)
        self.authorizations = AuthorizationsSyncModule(altscore_client)
        self.borrowers = BorrowersSyncModule(altscore_client)
        self.borrower_fields = BorrowerFieldsSyncModule(altscore_client)
        self.data_models = DataModelSyncModule(altscore_client)
        self.documents = DocumentsSyncModule(altscore_client)
        self.executions = ExecutionSyncModule(altscore_client)
        self.workflows = WorkflowsSyncModule(altscore_client)
        self.identities = IdentitiesSyncModule(altscore_client)
        self.points_of_contact = PointOfContactSyncModule(altscore_client)
        self.relationships = RelationshipsSyncModule(altscore_client)
        self.report_generator = ReportGeneratorSyncModule(altscore_client)
        self.form_templates = FormTemplatesSyncModule(altscore_client)
        self.otp_templates = OTPTemplatesSyncModule(altscore_client)
        self.store_packages = PackagesSyncModule(altscore_client)
        self.store_sources = SourcesSyncModule(altscore_client)
        self.evaluators = EvaluatorSyncModule(altscore_client)
        self.forms = FormsSyncModule(altscore_client)
        self.alerts = AlertsSyncModule(altscore_client)
        self.rules = RulesSyncModule(altscore_client)
        self.sat_integration = SatIntegrationSyncModule(altscore_client)



import datetime as dt
import httpx
from pydantic import BaseModel, Field
from altscore.common.http_errors import raise_for_status_improved, retry_on_401
from altscore.borrower_central.helpers import build_headers
from typing import Optional


class ExtractionCoverageInfo(BaseModel):
    rfc: str
    date_from: str = Field(alias="dateFrom")
    date_to: str = Field(alias="dateTo")
    has_coverage: bool = Field(alias="hasCoverage")
    has_running_extractions: bool = Field(alias="hasRunningExtractions")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class SatIntegrationAsyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client
        self.base_url = self.altscore_client._borrower_central_base_url

    def build_headers(self):
        return build_headers(self)

    async def check_extractions(
            self, rfc: str, date_to_analyze: Optional[dt.datetime] = None,
            days_of_tolerance: Optional[int] = None
    ) -> ExtractionCoverageInfo:
        payload = {
            "rfc": rfc
        }
        if date_to_analyze is not None:
            payload["dateToAnalyze"] = date_to_analyze.isoformat()
        if days_of_tolerance is not None:
            payload["daysOfTolerance"] = days_of_tolerance
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                "/v1/integrations/sat/extractions/check",
                json=payload,
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return ExtractionCoverageInfo.parse_obj(response.json())

    async def start_extractions(
            self, rfc: str, date_to_analyze: Optional[dt.datetime] = None
    ) -> None:
        payload = {
            "rfc": rfc
        }
        if date_to_analyze is not None:
            payload["dateToAnalyze"] = date_to_analyze.isoformat()

        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                "/v1/integrations/sat/extractions/start",
                json=payload,
                headers=self.build_headers()
            )
            raise_for_status_improved(response)
            return None


class SatIntegrationSyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client
        self.base_url = self.altscore_client._borrower_central_base_url

    def build_headers(self):
        return build_headers(self)

    def check_extractions(
            self, rfc: str, date_to_analyze: Optional[dt.datetime] = None,
            days_of_tolerance: Optional[int] = None
    ) -> ExtractionCoverageInfo:
        payload = {
            "rfc": rfc
        }
        if date_to_analyze is not None:
            payload["dateToAnalyze"] = date_to_analyze.isoformat()
        if days_of_tolerance is not None:
            payload["daysOfTolerance"] = days_of_tolerance

        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                "/v1/integrations/sat/extractions/check",
                json=payload,
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return ExtractionCoverageInfo.parse_obj(response.json())

    def start_extractions(
            self, rfc: str, date_to_analyze: Optional[dt.datetime] = None
    ) -> None:
        payload = {
            "rfc": rfc
        }
        if date_to_analyze is not None:
            payload["dateToAnalyze"] = date_to_analyze.isoformat()
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                "/v1/integrations/sat/extractions/start",
                json=payload,
                headers=self.build_headers()
            )
            raise_for_status_improved(response)
            return None



from typing import Optional, List, Dict, Any
import httpx
from altscore.altdata.model.data_request import RequestResult
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from pydantic import BaseModel, Field
import datetime as dt
from dateutil.parser import parse as parse_date


class PackageAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    source_id: Optional[str] = Field(alias="sourceId", default=None)
    content_type: Optional[str] = Field(alias="contentType", default=None)
    alias: Optional[str] = Field(alias="alias", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    tags: List[str] = Field(alias="tags")
    forced_stale: Optional[bool] = Field(alias="forcedStale", default=False)
    created_at: str = Field(alias="createdAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreatePackageDTO(BaseModel):
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    source_id: Optional[str] = Field(alias="sourceId", default=None)
    workflow_id: Optional[str] = Field(alias="workflowId", default=None)
    alias: Optional[str] = Field(alias="alias", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    content_type: Optional[str] = Field(alias="contentType", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    content: Any = Field(alias="content")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class PackageSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/stores/packages", header_builder, renew_token, PackageAPIDTO.parse_obj(data))


class PackageAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/stores/packages", header_builder, renew_token, PackageAPIDTO.parse_obj(data))


class PackagesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=PackageSync,
                         retrieve_data_model=PackageAPIDTO,
                         create_data_model=CreatePackageDTO,
                         update_data_model=None,
                         resource="stores/packages")

    def retrieve_package_by_alias(self, alias: str, data_age: Optional[dt.timedelta] = None) -> Optional[PackageSync]:
        packages = self.query(alias=alias, sort_by="createdAt", sort_order="desc")
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    def retrieve_source_package(
            self, source_id: str, borrower_id: Optional[str] = None, data_age: Optional[dt.timedelta] = None,
            package_alias: Optional[str] = None
    ) -> Optional[PackageSync]:
        if borrower_id:
            packages = self.query(
                borrower_id=borrower_id,
                source_id=source_id,
                alias=package_alias,
                sort_by="createdAt",
                sort_order="desc"
            )
        else:
            packages = self.query(
                source_id=source_id,
                alias=package_alias,
                sort_by="createdAt",
                sort_order="desc"
            )
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    def retrieve_workflow_package(
            self, workflow_id: str, alias: str, data_age: Optional[dt.timedelta] = None
    ) -> Optional[PackageSync]:
        packages = self.query(workflow_id=workflow_id, alias=alias, sort_by="createdAt", sort_order="desc")
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    def force_stale(self, package_id: Optional[str] = None, borrower_id: Optional[str] = None,
                    workflow_id: Optional[str] = None, alias: Optional[str] = None):
        if package_id is None and borrower_id is None and workflow_id is None and alias is None:
            raise ValueError("At least one of package_id, borrower_id, workflow_id or alias must be provided")
        body = {
            "packageId": package_id,
            "borrowerId": borrower_id,
            "workflowId": workflow_id,
            "alias": alias,
            "forcedStale": True
        }
        body = {k: v for k, v in body.items() if v is not None}
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            client.put(
                "/stores/packages/stale",
                json=body,
                headers=self.build_headers()
            )

    def create_from_altdata_request_result(
            self, borrower_id: str, source_id: str, altdata_request_result: RequestResult,
            attachments: Optional[List[Dict[str, Any]]] = None,
            content_type: str = "json", package_alias: Optional[str] = None
    ):
        package = altdata_request_result.to_package(source_id)
        bc_source_id = "AD_{}_{}".format(source_id, package["version"])
        package_data = {
            "borrower_id": borrower_id,
            "source_id": bc_source_id,
            "content": package,
            "content_type": content_type,
            "alias": package_alias
        }
        created_package_id = self.create(package_data)
        if attachments is not None:
            package_obj: PackageSync = self.retrieve(created_package_id)
            if package_obj is not None:
                for attachment in attachments:
                    package_obj.post_attachment(
                        attachment
                    )
        return created_package_id

    def create_all_from_altdata_request_result(
            self, borrower_id: str, altdata_request_result: RequestResult,
    ) -> Dict[str, str]:
        packages = {}
        for source_call_summary in altdata_request_result.call_summary:
            if source_call_summary.is_success:
                package_id = self.create_from_altdata_request_result(borrower_id=borrower_id,
                                                                     source_id=source_call_summary.source_id,
                                                                     altdata_request_result=altdata_request_result)
                packages[f"{source_call_summary.source_id}_{source_call_summary.version}"] = package_id
        return packages


class PackagesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=PackageAsync,
                         retrieve_data_model=PackageAPIDTO,
                         create_data_model=CreatePackageDTO,
                         update_data_model=None,
                         resource="/stores/packages")

    async def retrieve_package_by_alias(
            self, alias: str, data_age: Optional[dt.timedelta] = None
    ) -> Optional[PackageAsync]:
        packages = await self.query(alias=alias, sort_by="createdAt", sort_order="desc")
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    async def retrieve_source_package(
            self, source_id: str, borrower_id: Optional[str] = None, data_age: Optional[dt.timedelta] = None,
            package_alias: Optional[str] = None
    ) -> Optional[PackageAsync]:
        if borrower_id:
            packages = await self.query(borrower_id=borrower_id,
                                        source_id=source_id,
                                        sort_by="createdAt",
                                        sort_order="desc",
                                        alias=package_alias
                                        )
        else:
            packages = await self.query(
                source_id=source_id,
                sort_by="createdAt",
                sort_order="desc",
                alias=package_alias
            )
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    async def retrieve_workflow_package(
            self, workflow_id: str, alias: str, data_age: Optional[dt.timedelta] = None
    ) -> Optional[PackageAsync]:
        packages = await self.query(workflow_id=workflow_id, alias=alias, sort_by="createdAt", sort_order="desc")
        if len(packages) > 0:
            package = packages[0]
            if data_age is None:
                return package
            else:
                if parse_date(package.created_at) + data_age > dt.datetime.utcnow():
                    return package
        return None

    async def create_from_altdata_request_result(
            self, borrower_id: str, source_id: str, altdata_request_result: RequestResult,
            attachments: Optional[List[Dict[str, Any]]] = None, content_type: str = "json",
            package_alias: Optional[str] = None
    ):
        package = altdata_request_result.to_package(source_id)
        bc_source_id = "AD_{}_{}".format(source_id, package["version"])
        package_data = {
            "borrower_id": borrower_id,
            "source_id": bc_source_id,
            "content": package,
            "content_type": content_type,
            "alias": package_alias
        }
        created_package_id = await self.create(package_data)
        if attachments is not None:
            package_obj: PackageSync = await self.retrieve(created_package_id)
            if package_obj is not None:
                for attachment in attachments:
                    await package_obj.post_attachment(
                        attachment
                    )
        return created_package_id

    async def create_all_from_altdata_request_result(
            self, borrower_id: str, altdata_request_result: RequestResult,
    ) -> Dict[str, str]:
        packages = {}
        for source_call_summary in altdata_request_result.call_summary:
            if source_call_summary.is_success:
                package_id = await self.create_from_altdata_request_result(
                    borrower_id=borrower_id,
                    source_id=source_call_summary.source_id,
                    altdata_request_result=altdata_request_result)
                packages[f"{source_call_summary.source_id}_{source_call_summary.version}"] = package_id
        return packages



from pydantic import BaseModel, Field
from typing import Optional, List, Dict
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class SourceAPIDTO(BaseModel):
    id: str = Field(alias="id")
    label: Optional[str] = Field(alias="label")
    tags: List[str] = Field(alias="tags", default=[])
    content_type: str = Field(alias="contentType")
    transformer_config: Optional[Dict] = Field(alias="transformerConfig", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class TransformerConfig(BaseModel):
    url: str = Field(alias="url")
    headers: Dict[str, str] = Field(alias="headers")


class CreateSourceDTO(BaseModel):
    label: str = Field(alias="label")
    content_type: str = Field(alias="contentType")
    tags: List[str] = Field(alias="tags", default=[])
    transformer_config: Optional[TransformerConfig] = Field(alias="transformerConfig", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateAltdataSourceDTO(BaseModel):
    altdata_source_id: str = Field(alias="altdataSourceId")
    altdata_source_version: str = Field(alias="altdataSourceVersion")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class SourceSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/stores/sources", header_builder, renew_token, SourceAPIDTO.parse_obj(data))


class SourceAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/stores/sources", header_builder, renew_token, SourceAPIDTO.parse_obj(data))


class SourcesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=SourceSync,
                         retrieve_data_model=SourceAPIDTO,
                         create_data_model=CreateSourceDTO,
                         update_data_model=None,
                         resource="/stores/sources")

    def create_altdata(self, altdata_source_id: str, altdata_source_version: str) -> str:
        new_entity_data = {
            "altdataSourceId": altdata_source_id,
            "altdataSourceVersion": altdata_source_version
        }
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                f"/v1/stores/sources/altdata",
                headers=self.build_headers(),
                json=CreateAltdataSourceDTO.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json()["id"]


class SourcesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=SourceAsync,
                         retrieve_data_model=SourceAPIDTO,
                         create_data_model=CreateSourceDTO,
                         update_data_model=None,
                         resource="/stores/sources")

    async def create_altdata(self, altdata_source_id: str, altdata_source_version: str) -> str:
        new_entity_data = {
            "altdataSourceId": altdata_source_id,
            "altdataSourceVersion": altdata_source_version
        }
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/stores/sources/altdata",
                headers=self.build_headers(),
                json=CreateAltdataSourceDTO.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json()["id"]
